-------------------------------------------------------------------
Magic Toolscript---A Tcl-wrapped version of magic
-------------------------------------------------------------------

-------------------------------------------------------------------
Compiling and Installation
-------------------------------------------------------------------

Simplified methods for compiling and installing have not been introduced yet,
so follow these instructions.

1) You will need "swig" and Tcl version 8+.  Testing has been done on swig
   version 1.3.10, but lower versions should work okay.  Testing has been
   done on Tcl 8.3.4, but again, lower versions should suffice.

2) At the top level, do "make config" as usual.  DO NOT enable the modules
   for the Scheme embedded interpreter or the readline interface.  Neither
   of these work with Tcl.  Interactive features like "iroute" and "irsim"
   have not been tested, and may also not work.  BE SURE to enable
   threads.

3) Hand-edit the top-level Makefile, and add the following to the DFLAGS
   definition:
		-DMAGIC_WRAPPER

4) Do "make".  This will get through compiling all .c files and creating
   all libraries, but will halt with an error when generating the "magic"
   executable.

5) Do "make wrapper".  This will run "swig" and should complete the compilation 
   of the Tcl-compatible shared library file "magic/magicwrap.so".

6) Hand-edit magic/magic.sh and replace the HOME directory with the proper
   location of the source.  Note that this is supposed to fool tclsh into
   running the contents of ".tclshrc" from this directory on startup.  It
   is the only way to get tclsh to run both interactively and also process
   some commands.  On install, "magic/magic.sh" should go to
   /home/cad/bin/magic.sh or /home/cad/bin/magic if you intend to use it
   as the sole version of magic on the system; "magic/.tclshrc" should go
   into /home/cad/lib/magic/tcl/.tclshrc, and the HOME variable in "magic.sh"
   should then point to directory "/home/cad/lib/magic/tcl".

   The path to tclsh should be edited in magic.sh to match what is on the
   system.

7) ".tcl" files (packages) go in /home/cad/lib/magic/tcl/.

8) The startup script works out namespaces so that all magic commands
   which do not conflict with existing tcl commands become directly available
   from the command line.  Otherwise, all magic commands are prefixed by the 
   namespace "::magic::" to differentiate them from other applications.
   "pushing" the namespace is not possible in any real sense in Tcl (I don't
   know why; the need for it seems obvious enough), but the commands
   (defined in the .tclshrc startup file) "pushnamespace" and "popnamespace"
   do about 95% of it.

-------------------------------------------------------------------
Known Bugs
-------------------------------------------------------------------

1) Commands "list", "split", "array", "source", "flush", and "load"
   all conflict with Tcl commands.  Either my implementation of Tcl
   is broken, or else Tcl's "rename" command doesn't work as advertised,
   because what I want to do is to rename the unqualified Tcl command
   "list" as the magic command, and leave the qualified Tcl name "::list"
   as the Tcl version of list.  This would effectively act like a true
   namespace push/pop.  Instead, I do "rename list {}" and get
   "self-referential recursion in "unknown" for command "list""
   instead of "invalid command name "list"".

   Current implementation is to require that these command names be
   fully-qualified;  e.g., "::magic::load" or "magic::load".  The
   unqualified names (e.g., "::load" or "load") refer to the Tcl function.

2) The prompt is a bit quirky at times.  Two threads are producing the
   prompt, so sometimes it ends up with the wrong prompt, or two prompts
   (on startup) or no prompt at all (on quit).  At least some of these
   situations can be overcome with some careful coding of pthread
   conditional variables.

   The intended behavior is that the magic prompt ">" (or "]" while DRC
   is pending) appears when the cursor is in a window, indicating that
   the input is accepting keyboard macros.  When outside of the window
   or after using the macro ":" to get to the command-line entry from
   magic, the Tcl "%" prompt appears, indicating that all valid Tcl
   commands may be entered at the prompt.  Magic's prompt behavior has
   been altered to produce the extra space after the prompt rather than
   alter Tcl's prompt variable to ignore the space.

   Because terminal input and window input are two separate routines,
   it is not possible to enter part of a command in one and complete
   it in the other.  It should be possible for the "EnterNotify" and
   "LeaveNotify" event handlers to query what's already on the command
   line and pass it on to the correct destination.

3) The "help" command should either be reimplemented in Tcl commands,
   or else stdin needs to be reopened while the "more" ("less") command
   is running;  otherwise, the system locks up as the UNIX command waits
   for input from a closed stdin stream.

4) The "*" prefixing magic "wizard" commands is a bad idea and should be
   changed to something else.  While the "pushnamespace/popnamespace"
   commands deal with it properly, it is likely to be a perennial gotcha,
   and mishandling can cause atrocious things to happen.

5) This appears to be another Tcl bug, but if the same user is running
   Tcl in another window, magic will segmentation fault on startup.
   Have not tried to track this one down.

-------------------------------------------------------------------
