/* magicwrap.i --- SWIG directives for creating the interpreter-wrapped	*/
/* version of magic.							*/

%module magicwrap

%typemap(tcl8, ignore) int argc(int n_arg) {
    $1 = n_arg = objc - 1;
    objc = 2;
}

%typemap(tcl8, in) char **argv {
    int templength, i;
    $1 = (char **) mallocMagic(n_arg1 * sizeof(char *));
    for (i = 0; i < n_arg1; i++) {
	$1[i] = Tcl_GetStringFromObj(objv[i + 1], &templength);
    }
}

%typemap(tcl8, freearg) char **argv {
    freeMagic($1);
}

%{
#include <pthread.h>
#include <unistd.h>
#include <signal.h>

/* #include "magicwrap.h" */

#include "misc/magic.h"   
#include "utils/geometry.h"
#include "tiles/tile.h"  
#include "utils/hash.h"  
#include "utils/dqueue.h"
#include "database/database.h"
#include "windows/windows.h"
#include "commands/commands.h"
#include "utils/utils.h"
#include "textio/textio.h"
#include "textio/txcommands.h"
#include "utils/malloc.h" 

extern TxInputEvent txLastEvent;
extern DQueue *txInputCommands;
/*
 * String containing the version number of magic.  Don't change the string
 * here, nor its format.  It is updated by the Makefile in this directory.
 */

char *MagicVersion = MAGIC_VERSION;
char *MagicCompileTime = MAGIC_DATE;

int pipeIn, pipeOut;
Tcl_Interp *magicinterp;

/*--------------------------------------------------------------*/
/* Sends input from the interpreter to magic's TxDispatch loop  */
/*--------------------------------------------------------------*/

void
wrapperStdin ()
{
    int n, asize, alen;
    TxCommand *tclcmd;
    TxInputEvent *event = &txLastEvent;
 
    tclcmd = TxNewCommand();
    read(pipeIn, &(tclcmd->tx_argc), sizeof(int));

    alen = 0;
    for (n = 0; n < tclcmd->tx_argc; n++) {
        read(pipeIn, &asize, sizeof(int));
	read(pipeIn, &tclcmd->tx_argstring[alen], asize);
	tclcmd->tx_argv[n] = &tclcmd->tx_argstring[alen];
	alen += asize;
    }

    tclcmd->tx_p = event->txe_p;
    tclcmd->tx_wid = event->txe_wid;
 
    DQPushRear(txInputCommands, (ClientData)tclcmd);

    /* Add an event to trigger the callback */

    event = TxNewEvent();
    event->txe_button = TX_BYPASS;
    TxAddEvent(event);
}

/*--------------------------------------------------------------*/
/* Start TxDispatch, first closing its stdin so it can't get    */
/* data except through the interpreter.                         */
/*--------------------------------------------------------------*/
 
void
wrapperDispatch (nulldata)
    ClientData nulldata;
{
    TxDispatch((FILE *)NULL);

    /* NEVER REACHED */
}


/*----------------------------------------*/
/* Inline the two main wrapper prodedures */
/*----------------------------------------*/
 
static int _magic_start(ClientData clientData,
        Tcl_Interp *interp, int argc, char *argv[])
{
    pthread_t thread;
    int fildes[2];
    int status;
 
    /* Remember the interpreter */
    magicinterp = interp;

    fprintf(stderr, "Starting magic under Tcl interpreter\n");
    pipeOut = -1;
 
    mainInitBeforeArgs(argc, argv);
    mainDoArgs(argc, argv);
    mainInitAfterArgs();

    /* Prevent TxDispatch thread from receiving any input itself */
    /* (prevents magic from interpreting Tcl input as magic macros) */
 
    TxDelete1InputDevice(fileno(stdin));

    /* Add an input device to come from interpreter dispatcher */
 
    pipe(fildes);
    pipeIn = fildes[0];
    pipeOut = fildes[1];

    /* Save the stdin device for writing from magic window */

    TxAdd1InputDevice(pipeIn, wrapperStdin, (ClientData) NULL);
 
    /* Thread the dispatcher and fall through to Tcl interpreter */
 
    status = pthread_create(&thread, NULL, (void *)wrapperDispatch, (void *)NULL);

    TxResetTerminal();	/* Give original properties back to stdin */

    return status;
}

/*--------------------------------------------------------------*/
/* Dispatch a command from Tcl in the same way that the X loop  */
/* dispatches a command, via pipe.                              */
/*--------------------------------------------------------------*/
 
static int _tcl_dispatch(ClientData clientData,
        Tcl_Interp *interp, int argc, char *argv[]) {

    int n, asize;
 
    /* Diagnostic */
/*
    fprintf(stderr, "In _tcl_dispatch with command ");
    for (n = 0; n < argc; n++) {
        fprintf(stderr, "%s ", argv[n]);
    }
    fprintf(stderr, "\n");
*/

    /* File input commands that happen before magic's main input thread */
    /* is set up should append the input queue directly, because it's	*/
    /* all being handled in this thread.				*/

    if (pipeOut < 0)
    { 
	TxCommand *tclcmd;
 
	tclcmd = TxNewCommand();
	tclcmd->tx_argc = argc;

	asize = 0;
	for (n = 0; n < argc; n++) {
	    strcpy(&tclcmd->tx_argstring[asize], argv[n]);
	    tclcmd->tx_argv[n] = &tclcmd->tx_argstring[asize];
	    asize += (1 + strlen(argv[n]));
	}
	tclcmd->tx_wid = WIND_UNKNOWN_WINDOW;

	DQPushRear(txInputCommands, (ClientData)tclcmd);
    }
    else
    {
	/* Cause an event on the input by sending 1st char of 1st   */
	/* arg to the write pipe.                                   */
 
	write(pipeOut, &argc, sizeof(int));	/* trigger select() */
	for (n = 0; n < argc; n++)
	{
	    asize = 1 + strlen(argv[n]);
            write(pipeOut, (const void *)&asize, sizeof(int));
            write(pipeOut, (const void *)(argv[n]), asize);
	}
    }
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Send a command line which was collected by magic's TxEvent	*/
/* handler to the interpreter's event queue rather than sending	*/
/* it to TxParseString().					*/
/*--------------------------------------------------------------*/

void send_interpreter(char *line_in)
{
    char *reply;

    Tcl_Eval(magicinterp, line_in);

    reply = Tcl_GetStringResult(magicinterp);
    if (strlen(reply) > 0)
	fprintf(stderr, "%s\n", reply);
}

%}

%init %{
{
    extern char *windCommands[];
    int n;
    char keyword[100];
    char *kwptr = keyword + 7;

    /* Use namespace to avoid conflicts with existing commands */
    strcpy(keyword, "magic::");

    for (n = 0; CmdLongCommands[n] != NULL; n++)
    {
	sscanf(CmdLongCommands[n], "%s ", kwptr); /* get first word */
	Tcl_CreateCommand(interp, keyword, _tcl_dispatch,
			(ClientData)NULL, (Tcl_CmdDeleteProc *) NULL);
    }
    for (n = 0; windCommands[n] != NULL; n++)
    {
	sscanf(windCommands[n], "%s ", kwptr); /* get first word */
	Tcl_CreateCommand(interp, keyword, _tcl_dispatch,
			(ClientData)NULL, (Tcl_CmdDeleteProc *) NULL);
    }

    /* Start command */
    Tcl_CreateCommand(interp, "magic::start", _magic_start,
			(ClientData)NULL, (Tcl_CmdDeleteProc *) NULL);

    /* Export the namespace commands */

    Tcl_Eval(interp, "namespace eval magic namespace export *");
}
%}

/* %include "magicwrap.h" */
